


# from argparse import Namespace
# import json
# import logging
# import pathlib

# from pyschism import dates
# from pyschism.mesh import Hgrid, Vgrid, Fgrid

# logger = logging.getLogger(__name__)

# CONFIG_FILE_NAME = 'config.json'
# STATIC_DIRECTORY = 'static'
# FORECAST_DIRECTORY = 'forecast'


# class ForecastInit:

#     def __init__(self, args: Namespace):
#         self.args = args
#         self.save_user_arguments()

#     def save_user_arguments(self):
#         logger.info(
#             f"Writing configuration file to path {self.config_file}")
#         with open(self.config_file, 'w') as fp:
#             json.dump(self.args.__dict__, fp, indent=4)

#     @property
#     def project_directory(self):
#         if not hasattr(self, '_project_directory'):
#             self._project_directory = pathlib.Path(self.args.project_directory)
#             self._project_directory.mkdir(exist_ok=self.args.overwrite)
#         return self._project_directory

#     @property
#     def config_file(self):
#         return self.project_directory / CONFIG_FILE_NAME

#     @property
#     def static_directory(self):
#         if not hasattr(self, '_static_directory'):
#             self._static_directory = self.project_directory / STATIC_DIRECTORY
#             self._static_directory.mkdir(exist_ok=self.args.overwrite)
#         return self._static_directory

#     @property
#     def forecast_directory(self):
#         return self.project_directory / FORECAST_DIRECTORY

#     @property
#     def hgrid_path(self):
#         if not hasattr(self, '_hgrid_path'):
#             self._hgrid_path = self.static_directory / 'hgrid.gr3'
#             if not self._hgrid_path.exists() or self.args.overwrite is True:
#                 logger.info(
#                     f'Writing initial hgrid file to {self._hgrid_path}.')
#                 Hgrid.open(self.args.hgrid, crs=self.args.hgrid_crs).write(
#                         self._hgrid_path,
#                         overwrite=self.args.overwrite)
#         return self._hgrid_path

#     @property
#     def hgrid(self):
#         if not hasattr(self, '_hgrid'):
#             self._hgrid = Hgrid.open(self.hgrid_path, crs=self.args.hgrid_crs)
#         return self._hgrid

    # @property
    # def current_forecast_directory(self):
    #     return self.forecast_directory / dates.nearest_cycle_date()

    # @property
    # def start_date(self):


    

















# from argparse import Namespace
# from datetime import datetime, timedelta, timezone
# import json
# import logging
# import os
# import pathlib
# import shutil
# import subprocess

# import numpy as np  # type: ignore[import]
# import pytz

# from pyschism.domain import ModelDomain
# from pyschism.driver import ModelDriver
# from pyschism.forcing import Tides
# from pyschism.mesh.vgrid import Vgrid
# from pyschism.server import SlurmConfig

# # from pyschism.logger import get_logger


# class ProjectDirectory:

#     def __get__(self, obj, val):
#         project_directory = obj.__dict__.get('project_directory')
#         if project_directory is None:
#             project_directory = pathlib.Path(obj.args.project_directory)
#             project_directory.mkdir(parents=True, exist_ok=True)
#             obj.__dict__['project_directory'] = project_directory
#         return project_directory


# class ConfigFile:

#     def __get__(self, obj, val):
#         config_file = obj.__dict__.get('config_file')
#         if config_file is None:
#             config_file = obj.project_directory / 'config.json'
#             obj.__dict__['config_file'] = config_file
#         return config_file


# class ColdstartDirectory:

#     def __get__(self, obj, val):
#         coldstart_directory = obj.__dict__.get('coldstart_directory')
#         if coldstart_directory is None:
#             timestamp = str(obj.target_datetime).replace(' ', 'T')
#             coldstart_directory = obj.project_directory / \
#                 'coldstart' / f'{timestamp}'
#             coldstart_directory.mkdir(
#                 parents=True, exist_ok=True)
#             obj.__dict__['coldstart_directory'] = coldstart_directory
#         return coldstart_directory


# class ColdstartDomain:

#     def __get__(self, obj, val):
#         model_domain = obj.__dict__.get('coldstart_domain')
#         if model_domain is None:
#             model_domain = ModelDomain.open(
#                 obj.hgrid_path,
#                 obj.fgrid_path,
#                 # TODO: Vgrid.open() raises NotImplementedError
#                 # obj.vgrid_path,
#                 hgrid_crs=obj.args.hgrid_crs,
#                 fgrid_crs=obj.args.fgrid_crs
#             )
#             if obj.tides is not None:
#                 model_domain.add_boundary_condition(obj.tides)
#         return model_domain


# class ColdstartDriver:

#     def __get__(self, obj, val):
#         coldstart = obj.__dict__.get('coldstart_driver')
#         if coldstart is None:
#             start_date = obj.target_datetime - timedelta(
#                 days=obj.args.spinup_days)
#             rnday = obj.target_datetime - start_date
#             coldstart = ModelDriver(
#                     model_domain=obj.coldstart_domain,
#                     dt=obj.args.timestep,
#                     rnday=rnday,
#                     dramp=(2./3.)*rnday,
#                     start_date=start_date,
#                     nhot_write=True,
#                     # ibc=obj.ibc,
#                     # drampbc=obj.drampbc,
#                     server_config=obj.server_config,
#                     )
#             obj.__dict__['coldstart_driver'] = coldstart
#         return coldstart


# class HgridPath:

#     def __get__(self, obj, val):
#         hgrid_path = obj.__dict__.get('hgrid_path')
#         if hgrid_path is None:
#             hgrid_path = obj.static_files_directory / 'hgrid.gr3'
#             if hgrid_path.is_file() and obj.args.overwrite is True:
#                 hgrid_path.unlink()
#             if not hgrid_path.is_file():
#                 shutil.copy2(obj.args.hgrid, hgrid_path,
#                              follow_symlinks=True)
#             obj.__dict__['hgrid_path'] = hgrid_path
#         return hgrid_path


# class VgridPath:

#     def __get__(self, obj, val):
#         vgrid_path = obj.__dict__.get('vgrid_path')
#         if vgrid_path is None:
#             vgrid_path = obj.static_files_directory / 'vgrid.in'
#             if vgrid_path.is_file():
#                 if obj.args.overwrite is True:
#                     vgrid_path.unlink()
#             if obj.args.vgrid is None:
#                 if not vgrid_path.exists():
#                     Vgrid().write(vgrid_path)
#             else:
#                 shutil.copy2(
#                     obj.args.vgrid, vgrid_path, follow_symlinks=True)
#             obj.__dict__['vgrid_path'] = vgrid_path
#         return vgrid_path


# class FgridPath:

#     def __get__(self, obj, val):
#         fgrid_path = obj.__dict__.get('fgrid_path')
#         if fgrid_path is None:
#             if obj.args.fgrid_type == 'auto':
#                 fgrid = pathlib.Path(obj.args.fgrid)
#                 fgrid_path = obj.static_files_directory / fgrid.name
#             else:
#                 fgrid_path = obj.static_files_directory / \
#                     (obj.args.fgrid_type + '.gr3')
#             if obj.args.overwrite is True:
#                 if fgrid_path.is_file():
#                     os.remove(fgrid_path)
#             shutil.copy2(obj.args.fgrid, fgrid_path, follow_symlinks=True)
#             obj.__dict__['fgrid_path'] = fgrid_path
#         return fgrid_path


# class StaticFilesDirectory:

#     def __get__(self, obj, val):
#         static_files = obj.__dict__.get('static_files')
#         if static_files is None:
#             static_files = obj.project_directory / 'static'
#             static_files.mkdir(exist_ok=True)
#             obj.__dict__['static_files'] = static_files
#         return static_files


# class TidesDescriptor:

#     def __get__(self, obj, val):
#         tides = obj.__dict__.get('tides')
#         if tides is None:
#             if not obj.args.all_constituents \
#                     and not obj.args.major_constituents \
#                     and not obj.args.constituents:
#                 return
#             else:
#                 tides = Tides(
#                     database=obj.args.tidal_database,
#                     velocity=obj.args.bnd_vel
#                 )
#                 if obj.args.all_constituents:
#                     tides.use_all()
#                 if obj.args.major_constituents:
#                     tides.use_major()
#                 if obj.args.constituents:
#                     for constituent in obj.args.constituents:
#                         tides.use_constituent(constituent)
#                 if obj.args.Z0 is not None:
#                     tides.add_Z0(obj.args.Z0)
#             obj.__dict__['tides'] = tides
#         return tides


# class ServerConfigDescriptor:

#     def __get__(self, obj, val):
#         server_config = obj.__dict__.get("server_config")
#         if server_config is None:
#             if obj.args.server_config == "slurm":
#                 kwargs = {
#                     "account": obj.args.account,
#                     "ntasks": obj.args.nproc,
#                     "partition": obj.args.partition,
#                     "walltime": obj.args.walltime,
#                     "mail_type": obj.args.mail_type,
#                     "mail_user": obj.args.mail_user,
#                     "log_filename": obj.args.log_filename,
#                     "modules": obj.args.modules,
#                     # "schism_binary": obj.args.schism_binary,
#                     "extra_commands": obj.args.extra_commands,
#                     "launcher": obj.args.slurm_launcher,
#                     "nodes": obj.args.slurm_nodes
#                 }
#                 # if obj.args.slurm_filename is not None:
#                 #     kwargs.update({"filename": obj.args.ntasks})
#                 # if obj.args.slurm_rundir is not None:
#                 #     kwargs.update({"run_directory": obj.args.slurm_rundir})
#                 # if obj.args.run_name is not None:
#                 #     kwargs.update({"run_name": obj.args.run_name})
#                 server_config = SlurmConfig(**kwargs)
#                 obj.__dict__['server_config'] = server_config
#         return server_config


# class ForecastInit:

#     project_directory = ProjectDirectory()
#     config_file = ConfigFile()
#     coldstart_directory = ColdstartDirectory()
#     coldstart_domain = ColdstartDomain()
#     coldstart_driver = ColdstartDriver()
#     static_files_directory = StaticFilesDirectory()
#     hgrid_path = HgridPath()
#     vgrid_path = VgridPath()
#     fgrid_path = FgridPath()
#     tides = TidesDescriptor()
#     server_config = ServerConfigDescriptor()

#     def __init__(self, args: Namespace):

#         """
#         TODO: Write a file that inidicates this project is currently being
#         accessed by a process. Write the pid of this process to the file.
#         May be implemented by a "check_project()" method.
#         """
#         self._args = args
#         self._write_config_file()
#         self._symlink_files(self.coldstart_directory,
#                             self.coldstart_domain.ics)
#         self.logger.info('Writting coldstart files to disk...')
#         self.coldstart_driver.write(
#             self.coldstart_directory,
#             hgrid=False,
#             vgrid=False,
#             fgrid=False,
#             wind_rot=False,
#             overwrite=self.args.overwrite,
#             use_param_template=self.args.use_param_template
#         )

#         if self.args.skip_run is False:
#             # release memory before launching SCHISM.
#             self.logger.info('Releasing memory before calling SCHISM...')
#             for item in list(self.__dict__.keys()):
#                 if not item.startswith('_'):
#                     del self.__dict__[item]
#             self.logger.info('Calling SCHISM using make.')
#             subprocess.check_call(
#                 ["make", "run"],
#                 cwd=self.coldstart_directory
#             )
#         # else:
#             # self.logger.info("Skipping coldstart run.")

#         # self.logger.info("Finished coldstart sequence.")

#     @property
#     def args(self):
#         return self._args

#     def _write_config_file(self):
#         if self.config_file.is_file() and self.args.overwrite is False:
#             raise IOError(
#                 f'The given directory {str(self.project_directory)} has '
#                 'been initialized previously. Please use\npyschism '
#                 'forecast --overwrite init [...]\nto allow overwrite of '
#                 'previous initialization options.')
#         self.logger.info(
#             f"Writting configuration file to path {self.config_file}")
#         with open(self.config_file, 'w') as fp:
#             json.dump(self.args.__dict__, fp, indent=4)

#     def _symlink_files(self, target_directory, ics):
#         self.logger.info(
#             f"Establishing symlinks to target_directory: {target_directory}")
#         hgrid_lnk = target_directory / 'hgrid.gr3'
#         vgrid_lnk = target_directory / 'vgrid.in'
#         fgrid_lnk = target_directory / f'{self.fgrid_path.name}'
#         if self.args.overwrite is True:
#             if hgrid_lnk.exists():
#                 hgrid_lnk.unlink()
#             if vgrid_lnk.exists():
#                 vgrid_lnk.unlink()
#             if fgrid_lnk.exists():
#                 fgrid_lnk.unlink()
#         os.symlink(
#             os.path.relpath(
#                 self.hgrid_path, target_directory), hgrid_lnk)
#         os.symlink(
#             os.path.relpath(
#                 self.vgrid_path, target_directory), vgrid_lnk)
#         os.symlink(
#             os.path.relpath(
#                 self.fgrid_path, target_directory), fgrid_lnk)
#         if ics == 2:
#             self._symlink_hgridll(target_directory)

#     def _symlink_hgridll(self, target_directory):
#         hgridll_lnk = target_directory / 'hgrid.ll'
#         if hgridll_lnk.exists():
#             hgridll_lnk.unlink()
#         if self.args.overwrite is True:
#             if self.args.overwrite is True:
#                 if hgridll_lnk.exists():
#                     hgridll_lnk.unlink()
#         os.symlink(os.path.relpath(
#             self.hgrid_path, target_directory), hgridll_lnk)

#     @property
#     def target_datetime(self):
#         if not hasattr(self, '_target_datetime'):
#             now_utc = datetime.now(timezone.utc)
#             nearest_cycle = int(6 * np.floor(now_utc.hour/6))
#             self._target_datetime = datetime(
#                 now_utc.year, now_utc.month, now_utc.day, nearest_cycle,
#                 tzinfo=timezone.utc)
#             self.logger.info(
#                 f'Target datetime is: {str(self._target_datetime)}.')
#         return self._target_datetime

#     @property
#     def logger(self):
#         if not hasattr(self, '_logger'):
#             self._logger = logging.getLogger(f"{self.__class__.__name__}")
#         return self._logger



from collections import OrderedDict
from datetime import datetime, timedelta, timezone
from enum import Enum
import logging
# import os
from typing import Callable, Union

import numpy as np

# from pyschism.forcing.tides import bctypes
from pyschism.forcing.bctides.tpxo import TPXO
from pyschism.forcing.bctides.fes2014 import FES2014
from pyschism.forcing.bctides.hamtide import HAMTIDE
# from pyschism.forcing.baroclinic.gofs import GOFS
# from pyschism.forcing.baroclinic.rtofs import RTOFS

_logger = logging.getLogger(__name__)

MAJOR_CONSTITUENTS = ('Q1', 'O1', 'P1', 'K1', 'N2', 'M2', 'S2', 'K2')
MINOR_CONSTITUENTS = ('Mm', 'Mf', 'M4', 'MN4', 'MS4', '2N2', 'S1')
ALL_CONSTITUENTS = MAJOR_CONSTITUENTS + MINOR_CONSTITUENTS


class TidalDatabase(Enum):
    TPXO = TPXO
    FES2014 = FES2014 
    HAMTIDE = HAMTIDE

    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid tidal database.")


class ActiveConstituents:

    def __get__(self, obj, val):
        active_constituents = obj.__dict__.get('active_constituents')
        if active_constituents is None:
            active_constituents = OrderedDict()
            obj.__dict__['active_constituents'] = active_constituents
        return active_constituents

    def __set__(self, obj, val):
        return obj.__dict__['active_constituents']


class Tides:

    _active_constituents = ActiveConstituents()
    major_constituents = MAJOR_CONSTITUENTS
    minor_constituents = MINOR_CONSTITUENTS
    constituents = ALL_CONSTITUENTS
    _nodal_factors = {}
    _earth_equilibrium_arguments = {}
    _tidal_species_types = {}
    _amplitude_constants = {}
    _amplitudes = {
            'elevation': {},
            'velocity': {},
        }

    _phases = {
        'elevation': {},
        'velocity': {},
    }

    def __init__(
            self,
            tidal_database: Union[str, TidalDatabase] = TidalDatabase.TPXO,
            constituents='all',
    ):
        """Main class for requesting tidal boundary forcing for a SCHISM run.

        The user should add tidal requests to this class via it's methods.

        Args:
            elevation (optional): Apply elevation forcing to boundary, defaults
                to True.
            velocity (optional): Apply velocity forcing to boundary (currently
                disabled), defaults to False.
            database (optional): Tidal database to use in order to obtain
                boundary initial conditions, defaults to TidalDatabase.TPXO
        """
        self.tidal_database = tidal_database

        if isinstance(constituents, str):
            constituents = [constituents]

        constituents = list(constituents) if constituents is not None else []

        if 'all' in constituents and len(constituents) > 1:
            raise ValueError('When using all, must only pass one')

        elif 'major' in constituents and len(constituents) > 1:
            raise ValueError('When using major, must only pass one')

        if 'all' in constituents:
            self.use_all()

        elif 'major' in constituents:
            self.use_major()

        else:
            for constituent in constituents:
                self.use_constituent(constituent)

    def __iter__(self):
        for constituent in self.active_constituents:
            yield (constituent, self.get_tidal_constituent(constituent))

    def __len__(self):
        return len(self.active_constituents)

    def __call__(self, start_date: datetime, rnday: Union[int, datetime],
                 constituent: str):
        """Returns the tidal characteristic values for a given time period.
        """
        _logger.debug(
            f'Computing tidal factors for start_date={start_date} and '
            f'constituent {constituent}')
        return (self.get_tidal_species_type(constituent),
                self.get_tidal_potential_amplitude(constituent),  # TPK
                self.get_orbital_frequency(constituent),  # FF*
                self.get_nodal_factor(start_date, rnday, constituent),  # Amig*
                self.get_greenwich_factor(start_date, rnday, constituent))  # FACE* # noqa:E501

    def get_elevation(self, constituent, vertices):

        if constituent.lower() == 'z0':
            return np.full((vertices.shape[0],), float(self._Z0)), \
                    np.full((vertices.shape[0],), 0.)

        if constituent in self._amplitudes['elevation']:
            amps = self._amplitudes['elevation'][constituent]
            phases = self._phases['elevation'][constituent]
            if isinstance(amps, (float, int)):
                amps = np.full((vertices.shape[0],), float(amps))
            if isinstance(phases, (float, int)):
                phases = np.full((vertices.shape[0],), float(phases))
            return amps, phases
        
        # if vertices is None:
            # raise ValueError(
            #     'Argument vertices must not be None for tidal constituent '
            #     f'{constituent}.')
        return self.tidal_database.get_elevation(constituent, vertices)

    def get_velocity(self, constituent, vertices):
        if constituent.lower() == 'z0':
            return tuple(4*[np.full((vertices.shape[0],), 0.)])
        
        if constituent in self._amplitudes['velocity']:
            amps = self._amplitudes['velocity'][constituent]
            phases = self._phases['velocity'][constituent]
            if isinstance(amps, (float, int)):
                amps = np.full((vertices.shape[0],), float(amps))
            if isinstance(phases, (float, int)):
                phases = np.full((vertices.shape[0],), float(phases))
            return amps, phases

        # if vertices is None:
            # raise ValueError(
            #     'Argument vertices must not be None for tidal constituent '
            #     f'{constituent}.')
        return self.tidal_database.get_velocity(constituent, vertices)

    def use_all(self, potential=True, forcing=True):
        for constituent in self.tidal_database.constituents:
            if constituent not in self.tidal_potential_amplitudes:
                potential = False
            self._active_constituents[constituent] = {
                'potential': potential,
                'forcing': forcing
            }

    def use_major(self, potential=True, forcing=True):
        for constituent in self.tidal_database.constituents:
            if constituent in self.major_constituents:
                self.use_constituent(constituent, potential, forcing)

    def use_constituent(self, constituent, potential=True, forcing=True):
        if constituent not in self.constituents:
            raise ValueError(f"Constituent must be one of {self.constituents}")
        if constituent not in self.tidal_potential_amplitudes:
            potential = False
        self._active_constituents[constituent] = {
            "potential": potential,
            "forcing": forcing,
        }

    def drop_constituent(self, constituent):
        if constituent not in self.active_constituents:
            raise ValueError("Argument constituent must be one of "
                             f"{self.active_constituents}")
        self._active_constituents.pop(constituent)
        
    def add_constituent(
            self,
            name: str,
            angular_frequency,
            nodal_factor = 1.,
            earth_equilibrium_argument: float = 0.,
            amplitude_constant: float = None,
            tidal_species_type: int = None,
            elevation_amplitude: dict = None,
            elevation_phase: dict = 0.,
            velocity_amplitude: dict = None,
            velocity_phase = 0.,
    ):
        # self._amplitudes['elevation'].update({name: amplitude})
        self.orbital_frequencies.update({name: angular_frequency})
        # self.
        self._nodal_factors.update({name: nodal_factor})
        self._earth_equilibrium_arguments.update({name: earth_equilibrium_argument})
        self._tidal_species_types.update({name: tidal_species_type})
        self._amplitude_constants.update({name: amplitude_constant})  
        self._amplitudes['elevation'].update({name: elevation_amplitude})
        self._amplitudes['velocity'].update({name: velocity_amplitude})
        self._phases['elevation'].update({name: elevation_phase})
        self._phases['velocity'].update({name: velocity_phase})
        self._active_constituents[name] = {
                "potential": False,
                "forcing": True,
            }

    def add_Z0(self, value):
        self._active_constituents['Z0'] = {
            "potential": False,
            "forcing": True,
        }
        self._Z0 = float(value)

    def get_active_constituents(self):
        return list(self.active_constituents.keys())

    def get_active_forcing_constituents(self):
        return [const for const, _ in self.active_constituents.items()
                if _['forcing'] is True]

    def get_active_potential_constituents(self):
        return [const for const, _ in self.active_constituents.items()
                if _['potential'] is True]

    def get_tidal_potential_amplitude(self, constituent):
        if constituent in self.tidal_potential_amplitudes:
            return self.tidal_potential_amplitudes[constituent]

    def get_tidal_species_type(self, constituent):
        if constituent in self.tidal_species_type:
            return self.tidal_species_type[constituent]

    def get_orbital_frequency(self, constituent):
        # if constituent in self.
        return self.orbital_frequencies[constituent]

    def get_initial_conditions(self, constituent, vertices):
        return self.initial_conditions(constituent, vertices)

    def set_Z0(self, Z0):
        self._active_constituents['Z0'] = {
            'potential': False,
            'forcing': True
        }
        self._Z0 = Z0

    def _manage_dates(f: Callable):
        def decorator(self, start_date, rnday, constituent):
            val = f(self, start_date, rnday, constituent)
            if hasattr(self, 'start_date_utc'):
                del(self.start_date_utc)
            if hasattr(self, 'end_date_utc'):
                del(self.end_date_utc)
            return val
        return decorator

    @_manage_dates
    def get_nodal_factor(self, start_date: datetime,
                         rnday: Union[float, timedelta],
                         constituent: str):
        if constituent not in ALL_CONSTITUENTS:
            return self._nodal_factors[constituent]
        if start_date.tzinfo is not None and \
                start_date.tzinfo.utcoffset(start_date) is not None:
            self.start_date_utc = start_date.astimezone(timezone(timedelta(0)))
        else:
            self.start_date_utc = start_date

        if not isinstance(rnday, timedelta):
            rnday = timedelta(days=rnday)

        self.end_date_utc = self.start_date_utc + rnday
        if constituent == "M2":
            return self.EQ78
        elif constituent == "S2":
            return 1.0
        elif constituent == "N2":
            return self.EQ78
        elif constituent == "K1":
            return self.EQ227
        elif constituent == "M4":
            return (self.EQ78)**2.
        elif constituent == "O1":
            return self.EQ75
        elif constituent == "M6":
            return (self.EQ78)**3.
        elif constituent == "MK3":
            return self.EQ78*self.EQ227
        elif constituent == "S4":
            return 1.0
        elif constituent == "MN4":
            return (self.EQ78)**2.
        elif constituent == "Nu2":
            return self.EQ78
        elif constituent == "S6":
            return 1.0
        elif constituent == "MU2":
            return self.EQ78
        elif constituent == "2N2":
            return self.EQ78
        elif constituent == "OO1":
            return self.EQ77
        elif constituent == "lambda2":
            return self.EQ78
        elif constituent == "S1":
            return 1.0
        elif constituent == "M1":
            return self.EQ207
        elif constituent == "J1":
            return self.EQ76
        elif constituent == "Mm":
            return self.EQ73
        elif constituent == "Ssa":
            return 1.0
        elif constituent == "Sa":
            return 1.0
        elif constituent == "Msf":
            return self.EQ78
        elif constituent == "Mf":
            return self.EQ74
        elif constituent == "RHO":
            return self.EQ75
        elif constituent == "Q1":
            return self.EQ75
        elif constituent == "T2":
            return 1.0
        elif constituent == "R2":
            return 1.0
        elif constituent == "2Q1":
            return self.EQ75
        elif constituent == "P1":
            return 1.0
        elif constituent == "2SM2":
            return self.EQ78
        elif constituent == "M3":
            return self.EQ149
        elif constituent == "L2":
            return self.EQ215
        elif constituent == "2MK3":
            return self.EQ227*self.EQ78**2
        elif constituent == "K2":
            return self.EQ235
        elif constituent == "M8":
            return self.EQ78**4
        elif constituent == "MS4":
            return self.EQ78
        if constituent == "Z0":
            return 1.
        else:
            msg = f'Unrecognized constituent {constituent}'
            raise TypeError(msg)

    def _normalize_to_360(f: Callable):
        def decorator(self, start_date, rnday, constituent):
            return f(self, start_date, rnday, constituent) % 360.
        return decorator

    @_manage_dates
    @_normalize_to_360
    def get_greenwich_factor(self, start_date: datetime,
                             rnday: Union[float, timedelta],
                             constituent: str):
        if constituent in self._earth_equilibrium_arguments:
            return self._earth_equilibrium_arguments[constituent]
        if start_date.tzinfo is not None and \
                start_date.tzinfo.utcoffset(start_date) is not None:
            self.start_date_utc = start_date.astimezone(timezone(timedelta(0)))
        else:
            self.start_date_utc = start_date

        if not isinstance(rnday, timedelta):
            rnday = timedelta(days=rnday)

        self.end_date_utc = self.start_date_utc + rnday
        if constituent == "M2":
            return 2.*(self.DT-self.DS+self.DH)+2.*(self.DXI-self.DNU)
        elif constituent == "S2":
            return 2.*self.DT
        elif constituent == "N2":
            return 2.*(self.DT+self.DH)-3.*self.DS+self.DP+2. \
                * (self.DXI-self.DNU)
        elif constituent == "K1":
            return self.DT+self.DH-90.-self.DNUP
        elif constituent == "M4":
            return 4.*(self.DT-self.DS+self.DH)+4.*(self.DXI-self.DNU)
        elif constituent == "O1":
            return self.DT-2.*self.DS+self.DH+90.+2.*self.DXI-self.DNU
        elif constituent == "M6":
            return 6.*(self.DT-self.DS+self.DH)+6.*(self.DXI-self.DNU)
        elif constituent == "MK3":
            return 3.*(self.DT+self.DH)-2.*self.DS-90.+2.*(self.DXI-self.DNU) \
                - self.DNUP
        elif constituent == "S4":
            return 4.*self.DT
        elif constituent == "MN4":
            return 4.*(self.DT+self.DH)-5.*self.DS+self.DP+4.\
                * (self.DXI-self.DNU)
        elif constituent == "Nu2":
            return 2.*self.DT-3.*self.DS+4.*self.DH-self.DP+2. \
                * (self.DXI-self.DNU)
        elif constituent == "S6":
            return 6.*self.DT
        elif constituent == "MU2":
            return 2.*(self.DT+2.*(self.DH-self.DS))+2.*(self.DXI-self.DNU)
        elif constituent == "2N2":
            return 2.*(self.DT-2.*self.DS+self.DH+self.DP)+2. \
                * (self.DXI-self.DNU)
        elif constituent == "OO1":
            return self.DT+2.*self.DS+self.DH-90.-2.*self.DXI-self.DNU
        elif constituent == "lambda2":
            return 2.*self.DT-self.DS+self.DP+180.+2.*(self.DXI-self.DNU)
        elif constituent == "S1":
            return self.DT
        elif constituent == "M1":
            return self.DT-self.DS+self.DH-90.+self.DXI-self.DNU+self.DQ
        elif constituent == "J1":
            return self.DT+self.DS+self.DH-self.DP-90.-self.DNU
        elif constituent == "Mm":
            return self.DS-self.DP
        elif constituent == "Ssa":
            return 2.*self.DH
        elif constituent == "Sa":
            return self.DH
        elif constituent == "Msf":
            return 2.*(self.DS-self.DH)
        elif constituent == "Mf":
            return 2.*self.DS-2.*self.DXI
        elif constituent == "RHO":
            return self.DT+3.*(self.DH-self.DS)-self.DP+90.+2.\
                * self.DXI-self.DNU
        elif constituent == "Q1":
            return self.DT-3.*self.DS+self.DH+self.DP+90.+2.*self.DXI-self.DNU
        elif constituent == "T2":
            return 2.*self.DT-self.DH+self.DP1
        elif constituent == "R2":
            return 2.*self.DT+self.DH-self.DP1+180.
        elif constituent == "2Q1":
            return self.DT-4.*self.DS+self.DH+2.*self.DP+90.+2.*self.DXI \
                - self.DNU
        elif constituent == "P1":
            return self.DT-self.DH+90.
        elif constituent == "2SM2":
            return 2.*(self.DT+self.DS-self.DH)+2.*(self.DNU-self.DXI)
        elif constituent == "M3":
            return 3.*(self.DT-self.DS+self.DH)+3.*(self.DXI-self.DNU)
        elif constituent == "L2":
            return 2.*(self.DT+self.DH)-self.DS-self.DP+180.+2.\
                * (self.DXI-self.DNU)-self.DR
        elif constituent == "2MK3":
            return 3.*(self.DT+self.DH)-4.*self.DS+90.+4.*(self.DXI-self.DNU) \
                + self.DNUP
        elif constituent == "K2":
            return 2.*(self.DT+self.DH)-2.*self.DNUP2
        elif constituent == "M8":
            return 8.*(self.DT-self.DS+self.DH)+8.*(self.DXI-self.DNU)
        elif constituent == "MS4":
            return 2.*(2.*self.DT-self.DS+self.DH)+2.*(self.DXI-self.DNU)
        elif constituent == "Z0":
            return 0
        else:
            msg = f'Unrecognized constituent {constituent}'
            raise TypeError(msg)

    def get_lunar_node(self):
        return (259.1560564 - 19.328185764 * self.DYR - .0529539336 * self.DDAY
                - .0022064139 * self.hour_middle)

    def get_lunar_perigee(self):
        return (334.3837214 + 40.66246584 * self.DYR + .111404016 * self.DDAY
                + .004641834 * self.hour_middle)

    def get_lunar_mean_longitude(self):
        return (277.0256206 + 129.38482032 * self.DYR + 13.176396768
                * self.DDAY + .549016532 * self.start_date_utc.hour)

    def get_solar_perigee(self):
        return (281.2208569 + .01717836 * self.DYR + .000047064 * self.DDAY
                + .000001961 * self.start_date_utc.hour)

    def get_solar_mean_longitude(self):
        return (280.1895014 - .238724988 * self.DYR + .9856473288 * self.DDAY
                + .0410686387 * self.start_date_utc.hour)

    @property
    def EQ73(self):
        """ """
        return (2./3.-np.sin(self.I)**2)/.5021

    @property
    def EQ74(self):
        """ """
        return np.sin(self.I)**2/.1578

    @property
    def EQ75(self):
        """ """
        return np.sin(self.I)*np.cos(self.I/2.)**2/.37988

    @property
    def EQ76(self):
        """ """
        return np.sin(2.*self.I)/.7214

    @property
    def EQ77(self):
        """ """
        return np.sin(self.I)*np.sin(self.I/2.)**2/.0164

    @property
    def EQ78(self):
        """ """
        return (np.cos(self.I/2)**4)/.91544

    @property
    def EQ149(self):
        """ """
        return np.cos(self.I/2.)**6/.8758

    @property
    def EQ197(self):
        """ """
        return np.sqrt(2.310+1.435*np.cos(2.*(self.P - self.XI)))

    @property
    def EQ207(self):
        """ """
        return self.EQ75*self.EQ197

    @property
    def EQ213(self):
        """ """
        return np.sqrt(1.-12.*np.tan(self.I/2.)**2*np.cos(2.*self.P)
                       + 36.*np.tan(self.I/2.)**4)

    @property
    def EQ215(self):
        """ """
        return self.EQ78*self.EQ213

    @property
    def EQ227(self):
        """ """
        return np.sqrt(.8965*np.sin(2.*self.I)**2+.6001*np.sin(2.*self.I)
                       * np.cos(self.NU)+.1006)

    @property
    def EQ235(self):
        """ """
        return .001+np.sqrt(19.0444*np.sin(self.I)**4+2.7702*np.sin(self.I)**2
                            * np.cos(2.*self.NU)+.0981)

    @property
    def active_constituents(self):
        return self._active_constituents.copy()

    @property
    def all_constituents(self):
        if isinstance(self.tidal_database, HAMTIDE):
            return self.major_constituents
        elif isinstance(self.tidal_database, TPXO) or isinstance(self.tidal_database, FES2014):
            return (*self.major_constituents, *self.minor_constituents)
        else:
            raise NotImplementedError(
                'Unhandled forcing database instance of type '
                f'{type(self.tidal_database)}')

    orbital_frequencies = {
        'M4':      0.0002810378050173,
        'M6':      0.0004215567080107,
        'MK3':     0.0002134400613513,
        'S4':      0.0002908882086657,
        'MN4':     0.0002783986019952,
        'S6':      0.0004363323129986,
        'M3':      0.0002107783537630,
        '2MK3':    0.0002081166466594,
        'M8':      0.0005620756090649,
        'MS4':     0.0002859630068415,
        'M2':      0.0001405189025086,
        'S2':      0.0001454441043329,
        'N2':      0.0001378796994865,
        'Nu2':     0.0001382329037065,
        'MU2':     0.0001355937006844,
        '2N2':     0.0001352404964644,
        'lambda2': 0.0001428049013108,
        'T2':      0.0001452450073529,
        'R2':      0.0001456432013128,
        '2SM2':    0.0001503693061571,
        'L2':      0.0001431581055307,
        'K2':      0.0001458423172006,
        'K1':      0.0000729211583579,
        'O1':      0.0000675977441508,
        'OO1':     0.0000782445730498,
        'S1':      0.0000727220521664,
        'M1':      0.0000702594512543,
        'J1':      0.0000755603613800,
        'RHO':     0.0000653117453487,
        'Q1':      0.0000649585411287,
        '2Q1':     0.0000623193381066,
        'P1':      0.0000725229459750,
        'Mm':      0.0000026392030221,
        'Ssa':     0.0000003982128677,
        'Sa':      0.0000001991061914,
        'Msf':     0.0000049252018242,
        'Mf':      0.0000053234146919,
        'Z0':      0}

    tidal_potential_amplitudes = {
        'M2': 0.242334,
        'S2': 0.112841,
        'N2': 0.046398,
        'K2': 0.030704,
        'K1': 0.141565,
        'O1': 0.100514,
        'P1': 0.046843,
        'Q1': 0.019256,
        'Z0': 0}

    tidal_species_type = {
        'M2': 2,
        'S2': 2,
        'N2': 2,
        'K2': 2,
        'K1': 1,
        'O1': 1,
        'P1': 1,
        'Q1': 1,
        'Z0': 0}

    @property
    def hour_middle(self):
        return self.start_date_utc.hour + (
            (self.end_date_utc - self.start_date_utc).total_seconds()
            / 3600 / 2)

    @property
    def I(self):  # noqa:E741,E743
        return np.arccos(.9136949-.0356926*np.cos(self.N))

    @property
    def N(self):
        return np.deg2rad(self.DN)

    @property
    def DN(self):
        return self.get_lunar_node()

    @property
    def DYR(self):
        return self.start_date_utc.year - 1900.

    @property
    def DDAY(self):
        return (self.start_date_utc.timetuple().tm_yday
                + int((self.start_date_utc.year-1901.)/4.) - 1)

    @property
    def NU(self):
        return np.arcsin(.0897056*np.sin(self.N)/np.sin(self.I))

    @property
    def DT(self):
        return (180.+self.start_date_utc.hour*(360./24))

    @property
    def DS(self):
        return self.get_lunar_mean_longitude()

    @property
    def DP(self):
        return self.get_lunar_perigee()

    @property
    def P(self):
        return np.deg2rad(self.DP)

    @property
    def DH(self):
        return self.get_solar_mean_longitude()

    @property
    def DP1(self):
        return self.get_solar_perigee()  # HR

    @property
    def DNU(self):
        return np.rad2deg(self.NU)

    @property
    def XI(self):
        return self.N-2.*np.arctan(.64412*np.tan(self.N/2)) - self.NU

    @property
    def DXI(self):
        return np.rad2deg(self.XI)

    @property
    def NUP(self):
        return np.arctan(np.sin(self.NU) / (
            np.cos(self.NU)+.334766/np.sin(2.*self.I)))

    @property
    def DNUP(self):
        return np.rad2deg(self.NUP)

    @property
    def DPC(self):
        return self.DP - self.DXI

    @property
    def PC(self):
        return np.deg2rad(self.DPC)

    @property
    def R(self):
        return (np.arctan(np.sin(2.*self.PC)
                          / ((1./6.)*(1./np.tan(.5*self.I))**2 - np.cos(2.*self.PC))))

    @property
    def DR(self):
        return np.rad2deg(self.R)

    @property
    def NUP2(self):
        return (np.arctan(np.sin(2.*self.NU)
                          / (np.cos(2.*self.NU)+.0726184 / np.sin(self.I)**2))/2.)

    @property
    def DNUP2(self):
        return np.rad2deg(self.NUP2)

    @property
    def Q(self):
        return np.arctan2((5.*np.cos(self.I)-1.)*np.sin(self.PC),
                          (7.*np.cos(self.I)+1.)*np.cos(self.PC))

    @property
    def DQ(self):
        return np.rad2deg(self.Q)

    @property
    def ntip(self):
        return len(self.get_active_potential_constituents())

    @property
    def cutoff_depth(self):
        if self.ntip == 0:
            return 0
        try:
            return self.__cutoff_depth
        except AttributeError:
            return 50.

    @cutoff_depth.setter
    def cutoff_depth(self, cutoff_depth):
        assert isinstance(cutoff_depth, (int, float))
        self.__cutoff_depth = cutoff_depth

    @property
    def nbfr(self):
        return len(self.get_active_forcing_constituents())

    @property
    def tidal_database(self):
        return self._tidal_database

    @tidal_database.setter
    def tidal_database(self, database: str):
        if isinstance(database, str):
            database = TidalDatabase[database.upper()].value()
        if isinstance(database, TidalDatabase):
            database = database.value()
        self._tidal_database = database



from importlib import util

# from pyschism.param.param import Param
from pyschism.stations import Stations
from pyschism.driver import ModelDriver, ModelConfig

__all__ = ['ModelConfig', 'ModelDriver', 'Stations']


if util.find_spec("colored_traceback") is not None:
    import colored_traceback  # type: ignore[import]
    colored_traceback.add_hook(always=True)



from enum import Enum
import os
import pathlib
from typing import Union
import numpy as np

from pyproj import CRS  # type: ignore[import]
from shapely.geometry import Polygon, MultiPolygon, Point
import geopandas as gpd

from pyschism.mesh.base import Gr3
from pyschism.mesh.parsers import grd


class FrictionFilename(Enum):
    MANNINGS_N = 'manning.gr3'
    DRAG_COEFFICIENT = 'drag.gr3'
    ROUGHNESS_LENGTH = 'rough.gr3'

    @classmethod
    def _missing_(self, name):
        raise ValueError(f'{name} is not a valid filename for a friction '
                         'file.')


class NchiType(Enum):
    MANNINGS_N = -1
    ROUGHNESS_LENGTH = 1
    DRAG_COEFFICIENT = 0


class Fgrid(Gr3):
    """
    Base class for all friction types (e.g. manning.grd, drag.grd, etc...)
    """

    def __init__(self, nchi: NchiType, *argv, **kwargs):
        self._nchi = nchi
        self._fname = FrictionFilename[NchiType(nchi).name]
        super().__init__(*argv, **kwargs)

    @property
    def nchi(self):
        return self._nchi.value

    @property
    def fname(self):
        return self._fname.value

    @classmethod
    def open(cls, file: Union[str, os.PathLike],
             crs: Union[str, CRS] = None):
        filename = pathlib.Path(file).name
        if cls.__name__ == "Fgrid":
            return FrictionDispatch[
                FrictionFilename(filename).name].value(
                    **grd.read(pathlib.Path(file), boundaries=False, crs=crs))
        else:
            return super().open(file, crs)

    @classmethod
    def constant(cls, hgrid, value):
        obj = cls(**{k: v for k, v in hgrid.to_dict().items() if k
                     in ['nodes', 'elements', 'description', 'crs']})
        obj.values[:] = value
        obj.description = f'{cls.__name__.lower()} {obj.crs}'
        return obj

    def add_region(
            self,
            region: Union[Polygon, MultiPolygon],
            value
     ):
        # Assuming input polygons are in EPSG:4326
        if isinstance(region, Polygon):
            region = [region]
        gdf1 = gpd.GeoDataFrame(
                {'geometry': region}, crs=self.crs)

        points = [Point(*coord) for coord in self.coords]
        gdf2 = gpd.GeoDataFrame(
                 {'geometry': points, 'index': list(range(len(points)))},
                crs=self.crs)
        gdf_in = gpd.sjoin(gdf2, gdf1, op="within")
        picks = ([i.index for i in gdf_in.itertuples()])
        self.values[picks] = value

    def modify_by_region(self, hgrid, fname, value, depth1, flag):
        '''
        reset (flag==0) or add (flag==1) value to a region
        '''
        lines=[line.strip().split() for line in open(fname, 'r').readlines()]
        data=np.squeeze(np.array([lines[3:]])).astype('float')
        x=data[:,0]
        y=data[:,1]
        coords = list( zip(x, y))
        poly = Polygon(coords)

        # Assuming input polygons are in EPSG:4326
        #if isinstance(region, Polygon):
        #    region = [region]
        gdf1 = gpd.GeoDataFrame(
                {'geometry': [poly]}, crs=self.crs)

        points = [Point(*coord) for coord in self.coords]
        gdf2 = gpd.GeoDataFrame(
                 {'geometry': points, 'index': list(range(len(points)))},
                crs=self.crs)
        gdf_in = gpd.sjoin(gdf2, gdf1, op="within")
        picks = [i.index for i in gdf_in.itertuples()]
        if flag == 0:
            self.values[picks] = value
        else:
            picks2 = np.where(-hgrid.values > depth1)
            picks3 = np.intersect1d(picks, picks2)
            self.values[picks3] = self.values[picks3] + value

class ManningsN(Fgrid):
    """  Class for representing Manning's n values.  """

    def __init__(self, *argv, **kwargs):
        self.hmin_man = 1.
        super().__init__(NchiType.MANNINGS_N, *argv, **kwargs)

    @classmethod
    def linear_with_depth(
            cls,
            hgrid: Union[str, os.PathLike, Gr3],
            min_value: float = 0.02,
            max_value: float = 0.05,
            min_depth: float = None,
            max_depth: float = None):

        # Inspired by https://github.com/schism-dev/schism/blob/master/src/Utility/Pre-Processing/NWM/Manning/write_manning.py
        obj = cls.constant(hgrid, np.nan)
        min_depth = np.min(-hgrid.values) if min_depth is None \
            else float(min_depth)
        max_depth = np.max(-hgrid.values) if max_depth is None \
            else float(max_depth)

        values = (
                min_value + (-hgrid.values - min_depth)
                * (max_value - min_value) / (max_depth - min_depth))

        if min_value is not None:
            values[values < min_value] = min_value

        if max_value is not None:
            values[values > max_value] = max_value

        obj.values[:] = values

        return obj


class RoughnessLength(Fgrid):

    def __init__(self, *argv, **kwargs):
        self.dzb_min = 0.5
        self.dzb_decay = 0.
        super().__init__(NchiType.ROUGHNESS_LENGTH, *argv, **kwargs)


class DragCoefficient(Fgrid):

    def __init__(self, *argv, **kwargs):
        super().__init__(NchiType.DRAG_COEFFICIENT, *argv, **kwargs)

    @classmethod
    def linear_with_depth(
            cls,
            hgrid: Union[str, os.PathLike, Gr3],
            depth1: float = -1.0,  # Are depth1 and depth2 positive up or positive down?
            depth2: float = -3.0,
            bfric_river: float = 0.0025,
            bfric_land: float = 0.025
    ):

        obj = cls.constant(hgrid, np.nan)

        values = (bfric_river + (depth1 + hgrid.values) *
                  (bfric_land - bfric_river) / (depth1-depth2))
        values[values > bfric_land] = bfric_land
        values[values < bfric_river] = bfric_river

        obj.values[:] = values

        return obj


class FrictionDispatch(Enum):
    MANNINGS_N = ManningsN
    DRAG_COEFFICIENT = DragCoefficient
    ROUGHNESS_LENGTH = RoughnessLength



from datetime import datetime, timedelta
import os

import numpy as np

if __name__ == "__main__":
    '''
    This script is used to link sflux
    srcdir: data directory, should have subfolder named as "20220201", "20220202"
    dstdir: path of run/sflux directory
    '''

    srcdir = '/sciclone/schism10/lcui01/schism20/ICOGS/ICOGS3D/Forecast/sflux'
    dstdir = '/sciclone/schism10/lcui01/schism20/ICOGS/ICOGS3D/Forecast/run/sflux'

    vars=["air", "prc", "rad"]

    startdate = datetime(2022, 2, 1)
    rnday = timedelta(days=5)

    timevector = np.arange(startdate, rnday, timedelta(days=1)).astype(datetime)
    print(timevector)

    for i, date in enumerate(timevector):
        srcgfs = f"{srcdir}/{date.strftime('%Y%m%d')}/gfs_{date.strftime('%Y%m%d%H')}.nc"
        srchrrr = f"{srcdir}/{date.strftime('%Y%m%d')}/hrrr_{date.strftime('%Y%m%d%H')}.nc"

        for var in vars:

            dstgfs = f"{dstdir}/sflux_{var}_1.{str(i+1).zfill(4)}.nc"
            os.symlink(srcgfs, dstgfs)

            dsthrrr = f"{dstdir}/sflux_{var}_2.{str(i+1).zfill(4)}.nc"
            os.symlink(srchrrr, dsthrrr)



from datetime import timedelta
import logging
from typing import Union

from pyschism.enums import (
    # IofWetdryVariables,
    # IofZcorVariables,
    IofHydroVariables,
    IofDvdVariables,
    IofWwmVariables,
    IofGenVariables,
    IofAgeVariables,
    IofSedVariables,
    IofEcoVariables,
    IofIcmVariables,
    IofCosVariables,
    IofFibVariables,
    IofSed2dVariables,
    IofMarshVariables,
    IofIceVariables,
    IofAnaVariables,
    SchoutType
)


_logger = logging.getLogger(__name__)


class SurfaceOutputVars:

    def __init__(self):
        self._surface_output_vars = {
            'iof_hydro': [(var.value, i) for i, var
                          in enumerate(IofHydroVariables)],
            'iof_dvd': [(var.value, i) for i, var
                        in enumerate(IofDvdVariables)],
            'iof_wwm': [(var.value, i) for i, var
                        in enumerate(IofWwmVariables)],
            'iof_gen': [(var.value, i) for i, var
                        in enumerate(IofGenVariables)],
            'iof_age': [(var.value, i) for i, var
                        in enumerate(IofAgeVariables)],
            'iof_sed': [(var.value, i) for i, var
                        in enumerate(IofSedVariables)],
            'iof_eco': [(var.value, i) for i, var
                        in enumerate(IofEcoVariables)],
            'iof_icm': [(var.value, i) for i, var
                        in enumerate(IofIcmVariables)],
            'iof_cos': [(var.value, i) for i, var
                        in enumerate(IofCosVariables)],
            'iof_fib': [(var.value, i) for i, var
                        in enumerate(IofFibVariables)],
            'iof_sed2d': [(var.value, i) for i, var
                          in enumerate(IofSed2dVariables)],
            'iof_marsh': [(var.value, i) for i, var
                          in enumerate(IofMarshVariables)],
            'iof_ice': [(var.value, i) for i, var
                        in enumerate(IofIceVariables)],
            'iof_ana': [(var.value, i) for i, var
                        in enumerate(IofAnaVariables)],
        }

    def __get__(self, obj, val):
        return self._surface_output_vars


class OutputVariableDescriptor:

    def __init__(self, iof_type, name, index):
        self._iof_type = iof_type
        self._name = name
        self._index = index

    def __get__(self, obj, val):
        return bool(getattr(obj, f'_{self._iof_type}')[self._index])

    def __set__(self, obj, val: bool):
        if not isinstance(val, bool):
            raise TypeError(f'Argument to {self._name} must be boolean, not '
                            f'type {type(val)}.')
        iof = getattr(obj, f'_{self._iof_type}')
        iof[self._index] = int(val)


class SchoutMeta(type):

    surface_output_vars = SurfaceOutputVars()

    def __new__(meta, name, bases, attrs):
        for iof_type in meta.surface_output_vars.keys():
            attrs[f'_{iof_type}'] = len(SchoutType[iof_type].value)*[0]

        for iof_type, vardata in meta.surface_output_vars.items():
            for name, index in vardata:
                attrs[name] = OutputVariableDescriptor(iof_type, name, index)
        output_vars = []
        for iof_, outputs in meta.surface_output_vars.items():
            for name, id in outputs:
                output_vars.append(name)
        attrs['surface_output_vars'] = output_vars
        return type(name, bases, attrs)


class SCHOUT(
        metaclass=SchoutMeta
):
    """ Provides error checking implementation for SCHOUT group """

    def __init__(
            self,
            nhot_write: int = None,
            nspool_sta: int = None,
            **outputs
    ):
        """
        nhot_write:
            - if -1 will write last timestep (default)
            - if None it will be disabled.
            - if int interpreted as iteration
            - if timedelta it will be rounded to the nearest iteration
        """

        self.nhot_write = nhot_write
        self.nspool_sta = nspool_sta

        for key, val in outputs.items():
            setattr(self, key, val)

    def __iter__(self):
        for outvar in self._surface_output_vars:
            yield outvar, getattr(self, outvar)

    def __str__(self):
        schout = ["&SCHOUT"]
        if self.nhot_write is not None:
            schout.append(f"  nhot={self.nhot}")
            schout.append(f"  nhot_write={self.nhot_write}")
        if self.nspool_sta is not None:
            nspool_sta = self.nspool_sta
            if isinstance(nspool_sta, timedelta):
                nspool_sta = int(round(nspool_sta.total_seconds() / self.dt))
            if isinstance(nspool_sta, float):
                nspool_sta = int(
                    round(timedelta(hours=nspool_sta) / self.dt))
            if isinstance(nspool_sta, (int, float)):
                if nspool_sta <= 0:
                    raise ValueError("nspool_sta must be positive.")
            schout.append(f"  iout_sta={self.iout_sta}")
            schout.append(f"  nspool_sta={nspool_sta}")
        for var in dir(self):
            if var.startswith('iof'):
                for i, state in enumerate(getattr(self, var)):
                    if state == 1:
                        schout.append(f'  {var[1:]}({i+1})={state}')
        schout.append('/')
        return '\n'.join(schout)

    def to_dict(self):
        data = {}
        if self.nhot_write is not None:
            data['nhot'] = self.nhot
            data['nhot_write'] = self.nhot_write
        if self.nspool_sta is not None:
            nspool_sta = self.nspool_sta
            if isinstance(nspool_sta, timedelta):
                nspool_sta = int(round(nspool_sta.total_seconds() / self.dt))
            if isinstance(nspool_sta, float):
                nspool_sta = int(
                    round(timedelta(hours=nspool_sta) / self.dt))
            if isinstance(nspool_sta, (int, float)):
                if nspool_sta <= 0:
                    raise ValueError("nspool_sta must be positive.")
            data['iout_sta'] = self.iout_sta
            data['nspool_sta'] = nspool_sta
        for var in dir(self):
            if var.startswith('iof'):
                _var = var[1:]
                data[_var] = len(getattr(self, var)) * [0]
                for i, state in enumerate(getattr(self, var)):
                    if state == 1:
                        data[_var][i] = state
        return data

    @property
    def nhot_write(self):
        return self._nhot_write

    @nhot_write.setter
    def nhot_write(self, nhot_write: Union[int, None]):
        if nhot_write is not None:
            if not isinstance(nhot_write, int):
                raise TypeError(
                    f'Argument nhot_write must be of type {int} or None, not '
                    f'type {type(nhot_write)}.')
        self._nhot_write = nhot_write

    @property
    def nhot(self) -> Union[int, None]:
        if not hasattr(self, '_nhot') and self.nhot_write is not None:
            return 1
        else:
            return self._nhot

    @nhot.setter
    def nhot(self, nhot: Union[int, None]):
        if nhot not in [0, 1]:
            raise ValueError('Argument nhot must be 0, 1.')
        self._nhot = nhot

    @nhot.deleter
    def nhot(self):
        if hasattr(self, '_nhot'):
            del self._nhot

    @property
    def nspool_sta(self):
        return self._nspool_sta

    @nspool_sta.setter
    def nspool_sta(self, nspool_sta: Union[int, None]):
        if nspool_sta is not None:
            if not isinstance(nspool_sta, int):
                raise TypeError(
                    f'Argument nspool_sta must be of type {int} or None, not '
                    f'type {type(nspool_sta)}.')
        self._nspool_sta = nspool_sta

    @property
    def iout_sta(self) -> Union[int, None]:
        if not hasattr(self, '_iout_sta') and self.nspool_sta is not None:
            return 1
        else:
            return self._iout_sta

    @iout_sta.setter
    def iout_sta(self, iout_sta: Union[int, None]):
        if iout_sta not in [0, 1]:
            raise ValueError('Argument iout_sta must be 0, 1.')
        self._iout_sta = iout_sta

    @iout_sta.deleter
    def iout_sta(self):
        if hasattr(self, '_iout_sta'):
            del self._iout_sta



from abc import ABC, abstractmethod
import argparse


class CliComponent(ABC):

    @abstractmethod
    def __init__(self, args: argparse.Namespace):
        self.args = args

    @staticmethod
    @abstractmethod
    def add_subparser_action(subparsers: argparse._SubParsersAction) -> None:
        raise NotImplementedError('Method must be implemented by subclass.')



from pyschism.forcing.bctides.bctypes import Bctype


class Itrtype(Bctype):

    @property
    def itrtype(self) -> int:
        '''Returns integer representig SCHISM itrtype code for bctides.in'''

    @property
    def forcing_digit(self):
        return self.itrtype

    def get_boundary_string(self, hgrid, boundary):
        return ''


class UniformTimeHistoryTracer(Itrtype):

    def __init__(self, time_history):
        raise NotImplementedError(f'{self.__class__.__name__}')
        self.time_history = time_history

    @property
    def itrtype(self) -> int:
        return 1


class ConstantTracer(Itrtype):

    def __init__(self, value):
        raise NotImplementedError(f'{self.__class__.__name__}')
        self.value = value

    @property
    def itrtype(self) -> int:
        return 2


class TracerInitialConditions(Itrtype):

    def __init__(self):
        raise NotImplementedError(f'{self.__class__.__name__}')

    @property
    def itrtype(self):
        return 3


class SpatiallyVaryingTimeHistoryTracer(Itrtype):

    def __init__(self, data_source):
        self.data_source = data_source

    @property
    def itrtype(self):
        return 4


class Itrtype1(UniformTimeHistoryTracer):
    pass


class Itrtype2(ConstantTracer):
    pass


class Itrtype3(TracerInitialConditions):
    pass


class Itrtype4(SpatiallyVaryingTimeHistoryTracer):
    pass



#! /usr/bin/env python
import argparse
# from datetime import datetime
import logging
import os
import pathlib
# import shutil
import tarfile
import tempfile
import unittest
import urllib.request
from datetime import datetime, timedelta

from pyschism.cmd.bctides import BctidesCli
from pyschism.driver import ModelConfig
from pyschism.forcing.bctides import iettype, ifltype
from pyschism.mesh import Hgrid


logging.basicConfig(level=logging.INFO, force=True)


class ModelConfigurationTestCase(unittest.TestCase):

    def setUp(self):
        hgrid2D = os.getenv('NWM_TEST_MESH')
        if hgrid2D is None:
            data_directory = pathlib.Path(__file__).parent.absolute() / 'data'
            hgrid2D = data_directory / "GulfStreamDevel/hgrid.gr3"
            url = "https://www.dropbox.com/s/mjaxaqeggy721um/"
            url += "Gulf_Stream_develop.tar.gz?dl=1"
            g = urllib.request.urlopen(url)
            tmpfile = tempfile.NamedTemporaryFile()
            with open(tmpfile.name, 'b+w') as f:
                f.write(g.read())
            with tarfile.open(tmpfile.name, "r:gz") as tar:
                tar.extractall(data_directory / "GulfStreamDevel")
        self.hgrid2D = hgrid2D

    def _test_bctides_cli_2d(self):
        import tempfile
        tmpdir = tempfile.TemporaryDirectory()
        parser = argparse.ArgumentParser()
        add_bctides(parser.add_subparsers(dest='mode'))
        args = parser.parse_args(
            [
                'bctides',
                f'{self.hgrid2D}',  # hgrid path
                '--run-days=5',  # rndays
                '--hgrid-crs=epsg:4326',
                f'-o={tmpdir.name}',
                '--overwrite'
            ])
        BctidesCli(args)
        # with open(tmpdir.name + '/bctides.in') as f:
        #     for line in f:
        #         print(line, end='')

    def _test_bctides_cli_2d_w_velo(self):
        import tempfile
        tmpdir = tempfile.TemporaryDirectory()
        parser = argparse.ArgumentParser()
        add_bctides(parser.add_subparsers(dest='mode'))
        args = parser.parse_args(
            [
                'bctides',
                f'{self.hgrid2D}',  # hgrid path
                '--run-days=5',  # rndays,
                '--include-velocity',
                '--hgrid-crs=epsg:4326',
                f'-o={tmpdir.name}',
                '--overwrite'
            ])
        BctidesCli(args)
        # with open(tmpdir.name + '/bctides.in') as f:
        #     for line in f:
        #         print(line, end='')

    def test_bctides_cli_2d_w_elev2d(self):
        import tempfile
        tmpdir = tempfile.TemporaryDirectory()
        parser = argparse.ArgumentParser()
        add_bctides(parser.add_subparsers(dest='mode'))
        args = parser.parse_args(
            [
                'bctides',
                f'{self.hgrid2D}',  # hgrid path
                '--run-days=5',  # rndays,
                '--iettype-5',
                # '--ifltype-5',
                '--hgrid-crs=epsg:4326',
                f'-o={tmpdir.name}',
                '--overwrite'
            ])
        BctidesCli(args)

    def _test_bctides_cli_3d(self):
        import tempfile
        tmpdir = tempfile.TemporaryDirectory()
        parser = argparse.ArgumentParser()
        add_bctides(parser.add_subparsers(dest='mode'))
        args = parser.parse_args(
            [
                'bctides',
                f'{self.hgrid}',  # hgrid path
                '--run-days=5',  # rndays
                '--hgrid-crs=epsg:4326',
                f'-o={tmpdir.name}',
                '--overwrite'
            ])
        BctidesCli(args)

    def test_simple_tidal_setup(self):
        # Inputs: Modify As Needed!
        start_date = "2018091000" #in YYYYMMDDHH format
        end_date = "2018091600" #in YYYYMMDDHH format

        tide_source = "tpxo"

        dt = timedelta(seconds=150.)
        nspool = timedelta(minutes=20.)

        start_date = datetime.strptime(start_date, "%Y%m%d%H")
        end_date = datetime.strptime(end_date, "%Y%m%d%H")
        rnday = end_date - start_date

        hgrid = Hgrid.open(
            'https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll',
            crs=4326
        )

        config = ModelConfig(
            hgrid=hgrid,
            iettype=iettype.Iettype3(database=tide_source),
            ifltype=ifltype.Ifltype3(database=tide_source),
        )

        coldstart = config.coldstart(
            start_date=start_date,
            end_date=start_date + rnday,
            timestep=dt,
            nspool=timedelta(hours=1),
            elev=True,
            dahv=True,
        )

        with tempfile.TemporaryDirectory() as dn:
            tmpdir = pathlib.Path(dn)
            coldstart.write(tmpdir, overwrite=True)


if __name__ == '__main__':
    unittest.main()
